from abc import ABC, abstractmethod
from typing import Dict, Generic, Iterator, List, Optional, Sequence
from attr import define
from datamaestro.data import Base
from datamaestro.record import Record, Item
from datamaestro_text.data.ir import TopicRecord
from datamaestro_text.utils.iter import FactoryIterable, LazyList, RangeView

# ---- Basic types


class DecontextualizedItem(Item):
    """A topic record with decontextualized versions of the topic"""

    @abstractmethod
    def get_decontextualized_query(self, mode=None) -> str:
        """Returns the decontextualized query"""
        ...


@define
class SimpleDecontextualizedItem(DecontextualizedItem):
    """A topic record with one decontextualized version of the topic"""

    decontextualized_query: str

    @abstractmethod
    def get_decontextualized_query(self, mode=None) -> str:
        """Returns the decontextualized query"""
        assert mode is None

        return self.decontextualized_query


@define
class DecontextualizedDictItem(DecontextualizedItem):
    """A conversation entry providing decontextualized version of the user query"""

    default_decontextualized_key: str

    decontextualized_queries: Dict[str, str]

    def get_decontextualized_query(self, mode=None):
        return self.decontextualized_queries[mode or self.default_decontextualized_key]


class ConversationRecord(Record):
    """A conversation entry"""

    pass


class TopicConversationRecord(ConversationRecord, TopicRecord):
    """A conversation record"""

    pass


class AnswerConversationRecord(ConversationRecord):
    """A conversation record"""

    pass


@define
class AnswerEntry(Item):
    """A system answer"""

    answer: str
    """The system answer"""


@define
class RetrievedEntry(Item):
    """List of system-retrieved documents and their relevance"""

    documents: List[str]
    """List of retrieved documents"""

    document_relevances: Optional[List[str]] = None
    """List of retrieved documents and their relevance status"""


@define
class ClarifyingQuestionEntry(Item):
    """A system-generated clarifying question"""

    pass


#: The conversation
ConversationHistory = Sequence[ConversationRecord]


@define
class ConversationHistoryItem(Item):
    """A user interaction contextualized within a conversation"""

    history: ConversationHistory
    """The history"""


# ---- Abstract conversation representation


class ConversationNode:
    def entry(self) -> ConversationRecord:
        """The current conversation entry"""
        ...

    def history(self) -> ConversationHistory:
        """Preceding conversation entries, from most recent to more ancient"""
        ...


class ConversationTree:
    def __iter__(self) -> Iterator[ConversationNode]:
        """Iterates over conversation nodes"""
        pass


# ---- A conversation tree


class SingleConversationTree(ConversationTree):
    """Simple conversations, based on a sequence of entries"""

    id: str
    history: Sequence[ConversationRecord]

    def __init__(self, id: Optional[str], history: List[ConversationRecord]):
        """Create a simple conversation

        :param history: The entries, in reverse order (i.e. more ancient first)
        """
        self.history = history or []

    def add(self, entry: ConversationRecord):
        self.history.insert(0, entry)

    def __iter__(self) -> Iterator[ConversationNode]:
        for ix in range(len(self.history)):
            yield SingleConversationTreeNode(self, ix)


@define
class SingleConversationTreeNode(ConversationNode):
    tree: SingleConversationTree
    index: int

    def entry(self) -> ConversationRecord:
        return self.tree.history[self.index]

    def history(self) -> Sequence[ConversationRecord]:
        return self.tree.history[self.index + 1 :]


class ConversationTreeNode(ConversationNode, ConversationTree):
    """A conversation tree node"""

    entry: ConversationRecord
    parent: Optional["ConversationTreeNode"]
    children: List["ConversationTreeNode"]

    def __init__(self, entry):
        self.entry = entry
        self.parent = None
        self.children = []

    def add(self, node: "ConversationTreeNode") -> "ConversationTreeNode":
        self.children.append(node)
        node.parent = self
        return node

    def conversation(self, skip_self: bool) -> ConversationHistory:
        def iterator():
            current = self.parent if skip_self else self
            while current is not None:
                yield current.entry
                current = current.parent

        return LazyList(FactoryIterable(iterator))

    def __iter__(self) -> Iterator["ConversationTreeNode"]:
        """Iterates over all conversation tree nodes (pre-order)"""
        yield self.entry
        for child in self.children:
            yield from child


class ConversationDataset(Base, ABC):
    """A dataset made of conversations"""

    @abstractmethod
    def __iter__(self) -> Iterator[ConversationTree]:
        """Return an iterator over conversations"""
        for i in range(len(self)):
            return self.get(i)



from abc import ABC, abstractmethod
from attrs import define
from typing import List
from datamaestro.record import Record, Item, recordtypes


class BaseRecord(Record):
    @classmethod
    def from_text(cls, text: str, *items: Item):
        return cls(SimpleTextItem(text), *items)

    @classmethod
    def from_id(cls, id: str, *items: Item):
        return cls(IDItem(id), *items)


class TopicRecord(BaseRecord):
    """Topic record"""


class DocumentRecord(BaseRecord):
    """Document record"""


@define()
class ScoredItem(Item):
    """A score associated with the document"""

    score: float
    """A retrieval score associated with this record (e.g. of the first-stage
    retriever)"""


class TextItem(Item, ABC):
    @property
    @abstractmethod
    def text(self) -> str:
        """Returns the text"""


@define
class SimpleTextItem(TextItem):
    """A topic/document with a text record"""

    text: str


@define
class InternalIDItem(Item, ABC):
    """A topic/document with an internal ID"""

    id: int


@define
class IDItem(Item, ABC):
    """A topic/document with an external ID"""

    id: str


@define
class AdhocAssessment:
    doc_id: str


@define
class SimpleAdhocAssessment(AdhocAssessment):
    rel: float
    """Relevance (> 0 if relevant)"""


@define
class AdhocAssessedTopic:
    topic_id: str
    """The topic ID"""

    assessments: List[AdhocAssessment]
    """List of assessments for this topic"""


# --- Commonly used types


@recordtypes(IDItem)
class IDTopicRecord(TopicRecord):
    pass


@recordtypes(IDItem)
class IDDocumentRecord(DocumentRecord):
    pass


@recordtypes(SimpleTextItem)
class SimpleTextTopicRecord(TopicRecord):
    pass


@recordtypes(SimpleTextItem)
class SimpleTextDocumentRecord(DocumentRecord):
    pass


@recordtypes(IDItem, TextItem)
class GenericDocumentRecord(DocumentRecord):
    @classmethod
    def create(cls, id: str, text: str, *items: Item):
        return cls(IDItem(id), SimpleTextItem(text), *items)


@recordtypes(IDItem, TextItem)
class GenericTopicRecord(TopicRecord):
    @classmethod
    def create(cls, id: str, text: str, *items: Item):
        return cls(IDItem(id), SimpleTextItem(text), *items)



from typing import Callable, TypeVar, Iterator, List, Union

T = TypeVar("T")


class BatchIterator(Iterator[List[T]]):
    """Adapter for iterators to return batches of elements instead of"""

    def __init__(self, iterator: Iterator[T], size: int):
        self.iterator = iterator
        self.size = size

    def __next__(self):
        batch = []
        for _, element in zip(range(self.size), self.iterator):
            batch.append(element)

        if len(batch) == 0:
            raise StopIteration()

        return batch


class FactoryIterable:
    def __init__(self, factory: Callable[[], Iterator[T]]):
        self.factory: Callable[[], Iterator[T]] = factory

    def __iter__(self) -> Iterator[T]:
        return self.factory()


class RangeView:
    def __init__(self, source, key: Union[slice, range]):
        self.key = (
            key if isinstance(key, range) else range(key.start, key.stop, key.step)
        )
        self.source = source

    def __getitem__(self, key: Union[slice, int]):
        slice = self.range[key]

        if isinstance(slice, int):
            return self.source[slice]

        return RangeView(self.source, key)


class LazyList:
    """Iterable-based list

    The list is only materialized if needed"""

    def __init__(self, iterable, materialize_threshold=0):
        self.iterable = iterable
        self.materialized_list = None
        self.materialize_threshold = materialize_threshold

    def __iter__(self):
        # If not materialized, return an iterator over the iterable
        if self.materialized_list is None:
            return iter(self.iterable)
        # If materialized, return an iterator over the materialized list
        else:
            return iter(self.materialized_list)

    def __getitem__(self, index):
        # Materialize the list if accessing an index above the threshold or any slice
        if isinstance(index, slice) or index >= self.materialize_threshold:
            self._materialize()
        elif self.materialized_list is None:
            # Returns the element at the given index
            return [
                element for _, element in zip(range(index + 1), iter(self.iterable))
            ][-1]

        return self.materialized_list[index]

    def _materialize(self):
        # Convert the iterable to a list if it hasn't been already
        if self.materialized_list is None:
            self.materialized_list = list(self.iterable)



from abc import ABC, abstractmethod
from functools import partial
import logging
from pathlib import Path
from typing import Iterator, Tuple, Type, List
import ir_datasets
from ir_datasets.indices import PickleLz4FullStore
from ir_datasets.formats import (
    GenericDoc,
    GenericQuery,
    GenericDocPair,
    TrecParsedDoc,
    TrecQuery,
)
import ir_datasets.datasets as _irds
from experimaestro import Config, Param
from experimaestro.compat import cached_property
from experimaestro import Option
from datamaestro.record import recordtypes
import datamaestro_text.data.ir as ir
from datamaestro_text.data.ir.base import (
    Record,
    TopicRecord,
    DocumentRecord,
    SimpleTextItem,
    AdhocAssessedTopic,
    SimpleAdhocAssessment,
    IDItem,
    IDTopicRecord,
    IDDocumentRecord,
)
import datamaestro_text.data.ir.formats as formats


# Interface between ir_datasets and datamaestro:
# provides adapted data types


class IRDSId(Config):
    irds: Option[str]
    """The id to load the dataset from ir_datasets"""

    @classmethod
    def __xpmid__(cls):
        return f"ir_datasets.{cls.__qualname__}"

    @cached_property
    def dataset(self):
        return ir_datasets.load(self.irds)

    def iter(self) -> Iterator[Record]:
        """Returns an iterator over topics"""
        for query in self.dataset.queries_iter():
            yield self.factory(query)

    def count(self):
        return self.dataset.queries_count()


class AdhocAssessments(ir.AdhocAssessments, IRDSId):
    def iter(self):
        """Returns an iterator over assessments"""
        ds = self.dataset

        class Qrels(dict):
            def __missing__(self, key):
                qqrel = AdhocAssessedTopic(key, [])
                self[key] = qqrel
                return qqrel

        qrels = Qrels()
        for qrel in ds.qrels_iter():
            qrels[qrel.query_id].assessments.append(
                SimpleAdhocAssessment(qrel.doc_id, qrel.relevance)
            )

        return qrels.values()


class tuple_constructor:
    def __init__(self, target_cls: Type, id_field: str, *fields: str):
        self.id_field = id_field
        self.target_cls = target_cls
        self.fields = fields

    def check(self, source_cls: Type):
        source_fields = tuple(f for f in source_cls._fields if f != self.id_field)
        assert source_fields == self.fields, (
            "Internal error: Fields do not match, "
            f"source({source_cls.__qualname__})={source_fields}/{self.id_field} [vs] target={self.fields}"
        )

    def __call__(self, recordtype, entry):
        values = tuple(getattr(entry, key) for key in self.fields)
        return recordtype(
            IDItem(getattr(entry, self.id_field)), self.target_cls(*values)
        )


class Documents(ir.DocumentStore, IRDSId):
    CONVERTERS = {
        GenericDoc: tuple_constructor(SimpleTextItem, "doc_id", "text"),
        _irds.beir.BeirCordDoc: tuple_constructor(
            formats.CordDocument, "doc_id", "text", "title", "url", "pubmed_id"
        ),
        _irds.beir.BeirTitleDoc: tuple_constructor(
            formats.TitleDocument, "doc_id", "text", "title"
        ),
        _irds.beir.BeirTitleUrlDoc: tuple_constructor(
            formats.TitleUrlDocument, "doc_id", "text", "title", "url"
        ),
        _irds.msmarco_document.MsMarcoDocument: tuple_constructor(
            formats.MsMarcoDocument, "doc_id", "url", "title", "body"
        ),
        _irds.cord19.Cord19FullTextDoc: tuple_constructor(
            formats.CordFullTextDocument,
            "doc_id",
            "title",
            "doi",
            "date",
            "abstract",
            "body",
        ),
        _irds.nfcorpus.NfCorpusDoc: tuple_constructor(
            formats.NFCorpusDocument, "doc_id", "url", "title", "abstract"
        ),
        TrecParsedDoc: tuple_constructor(
            formats.TrecParsedDocument, "doc_id", "title", "body", "marked_up_doc"
        ),
        _irds.wapo.WapoDoc: tuple_constructor(
            formats.WapoDocument,
            "doc_id",
            "url",
            "title",
            "author",
            "published_date",
            "kicker",
            "body",
            "body_paras_html",
            "body_media",
        ),
        _irds.tweets2013_ia.TweetDoc: tuple_constructor(
            formats.TweetDoc,
            "doc_id",
            "text",
            "user_id",
            "created_at",
            "lang",
            "reply_doc_id",
            "retweet_doc_id",
            "source",
            "source_content_type",
        ),
    }

    """Wraps an ir datasets collection -- and provide a default text
    value depending on the collection itself"""

    # List of fields
    # self.dataset.docs_cls()._fields

    def __getstate__(self):
        return (self.id, self.irds)

    def __setstate__(self, state):
        self.id, self.irds = state

    def iter(self) -> Iterator[ir.DocumentRecord]:
        """Returns an iterator over adhoc documents"""
        for doc in self.dataset.docs_iter():
            yield self.converter(self.document_recordtype, doc)

    @property
    def documentcount(self):
        return self.dataset.docs_count()

    @cached_property
    def store(self):
        return self.dataset.docs_store()

    @cached_property
    def _docs(self):
        return self.dataset.docs_iter()

    def docid_internal2external(self, ix: int):
        return self._docs[ix].doc_id

    def document_ext(self, docid: str) -> DocumentRecord:
        return self.converter(self.document_recordtype, self.store.get(docid))

    def documents_ext(self, docids: List[str]) -> DocumentRecord:
        """Returns documents given their external IDs (optimized for batch)"""
        retrieved = self.store.get_many(docids)
        return [
            self.converter(self.document_recordtype, retrieved[docid])
            for docid in docids
        ]

    def document_int(self, ix):
        return self.converter(self.document_recordtype, self._docs[ix])

    @cached_property
    def document_recordtype(self):
        return DocumentRecord.from_types(
            f"{self.converter.target_cls.__name__}Record",
            IDItem,
            self.converter.target_cls,
            module=__name__,
        )

    @cached_property
    def converter(self):
        converter = Documents.CONVERTERS[self.dataset.docs_cls()]
        converter.check(self.dataset.docs_cls())
        return converter


if hasattr(_irds, "miracl"):
    Documents.CONVERTERS[_irds.miracl.MiraclDoc] = tuple_constructor(
        formats.DocumentWithTitle, "doc_id", "text", "title"
    )


# Fix while PR https://github.com/allenai/ir_datasets/pull/252
# is not in.
class DMPickleLz4FullStore(PickleLz4FullStore):
    def get_many(self, doc_ids, field=None):
        result = {}
        field_idx = self._doc_cls._fields.index(field) if field is not None else None
        for doc in self.get_many_iter(doc_ids):
            if field is not None:
                result[getattr(doc, self._id_field)] = doc[field_idx]
            else:
                result[getattr(doc, self._id_field)] = doc
        return result


class LZ4DocumentStore(ir.DocumentStore):
    """A LZ4-based document store"""

    path: Param[Path]

    #: Lookup field
    lookup_field: Param[str]

    # Extra indexed fields (e.g. URLs)
    index_fields: List[str]

    @cached_property
    def store(self):
        return DMPickleLz4FullStore(
            self.path, None, self.data_cls, self.lookup_field, self.index_fields
        )

    @cached_property
    def _docs(self):
        return self.store.__iter__()

    def docid_internal2external(self, ix: int):
        return getattr(self._docs[ix], self.store._id_field)

    def document_ext(self, docid: str) -> DocumentRecord:
        return self.converter(self.document_recordtype, self.store.get(docid))

    def documents_ext(self, docids: List[str]) -> DocumentRecord:
        """Returns documents given their external IDs (optimized for batch)"""
        retrieved = self.store.get_many(docids)
        return [
            self.converter(self.document_recordtype, retrieved[docid])
            for docid in docids
        ]

    def converter(self, data):
        """Converts a document from LZ4 tuples to any other format"""
        # By default, use identity
        return data

    def iter(self) -> Iterator[DocumentRecord]:
        """Returns an iterator over documents"""
        return map(
            partial(self.converter, self.document_recordtype), self.store.__iter__()
        )

    def documentcount(self):
        if self.count:
            return self.count
        return self.store.count()


class TopicsHandler(ABC):
    @abstractmethod
    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        ...

    @abstractmethod
    def topic_ext(self, external_topic_id: str) -> TopicRecord:
        """Returns a document given its external ID"""
        ...

    @abstractmethod
    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        ...


class SimpleTopicsHandler(TopicsHandler):
    def __init__(self, converter, topics: "Topics"):
        self.converter = converter
        self._topics = topics
        self.target_cls = converter.target_cls
        converter.check(topics.dataset.queries_cls())

    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        return self.topics_list[internal_topic_id]

    def topic_ext(self, external_topic_id: int) -> TopicRecord:
        """Returns a document given its external ID"""
        return self.topics_map[external_topic_id]

    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        yield from self.topics_list

    @cached_property
    def topics_map(self):
        return self.topics[0]

    @cached_property
    def topics_list(self):
        return self.topics[1]

    @cached_property
    def topics(self):
        topic_map = {}
        topic_list = []
        for query in self._topics.dataset.queries_iter():
            record = self.converter(self._topics.topic_recordtype, query)
            topic_map[query.query_id] = record
            topic_list.append(record)

        return topic_map, topic_list


class Topics(ir.TopicsStore, IRDSId):
    CONVERTERS = {
        GenericQuery: tuple_constructor(SimpleTextItem, "query_id", "text"),
        _irds.beir.BeirCovidQuery: tuple_constructor(
            formats.TrecTopic, "query_id", "text", "query", "narrative"
        ),
        _irds.beir.BeirUrlQuery: tuple_constructor(
            formats.UrlTopic, "query_id", "text", "url"
        ),
        _irds.nfcorpus.NfCorpusQuery: tuple_constructor(
            formats.NFCorpusTopic, "query_id", "title", "all"
        ),
        TrecQuery: tuple_constructor(
            formats.TrecTopic, "query_id", "title", "description", "narrative"
        ),
        _irds.tweets2013_ia.TrecMb13Query: tuple_constructor(
            formats.TrecMb13Query, "query_id", "query", "time", "tweet_time"
        ),
        _irds.tweets2013_ia.TrecMb14Query: tuple_constructor(
            formats.TrecMb14Query,
            "query_id",
            "query",
            "time",
            "tweet_time",
            "description",
        ),
    }

    HANDLERS = {
        cls: partial(SimpleTopicsHandler, converter)
        for cls, converter in CONVERTERS.items()
    }

    def count(self):
        return self.dataset.queries_count()

    @cached_property
    def topic_recordtype(self) -> Type[Record]:
        return TopicRecord.from_types(
            f"{self.handler.target_cls.__name__}Record",
            IDItem,
            self.handler.target_cls,
            module=Topics.__class__.__module__,
        )

    @cached_property
    def handler(self):
        handler = Topics.HANDLERS[self.dataset.queries_cls()](self)
        return handler

    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        return self.handler.topic_int(internal_topic_id)

    def topic_ext(self, external_topic_id: str) -> TopicRecord:
        """Returns a document given its external ID"""
        return self.handler.topic_ext(external_topic_id)

    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        return self.handler.iter()


if hasattr(_irds.trec_cast, "Cast2022Query"):
    from datamaestro_text.data.conversation.base import (
        ConversationTreeNode,
        DecontextualizedDictItem,
        RetrievedEntry,
        AnswerConversationRecord,
        ConversationHistoryItem,
    )

    class CastTopicsHandler(TopicsHandler):
        def __init__(self, dataset):
            self.dataset = dataset

        @property
        @abstractmethod
        def records(self):
            ...

        @cached_property
        def ext2records(self):
            return {record[IDItem].id: record for record in self.records}

        def topic_int(self, internal_topic_id: int) -> TopicRecord:
            """Returns a document given its internal ID"""
            return self.records[internal_topic_id]

        def topic_ext(self, external_topic_id: str) -> TopicRecord:
            """Returns a document given its external ID"""
            return self.ext2records[external_topic_id]

        def iter(self) -> Iterator[ir.TopicRecord]:
            """Returns an iterator over topics"""
            return iter(self.records)

    @recordtypes(
        IDItem, SimpleTextItem, DecontextualizedDictItem, ConversationHistoryItem
    )
    class Cast2020TopicRecord(TopicRecord):
        ...

    @recordtypes(RetrievedEntry)
    class Cast2020ResponseRecord(AnswerConversationRecord):
        ...

    class Cast2020TopicsHandler(CastTopicsHandler):
        @cached_property
        def records(self):
            try:
                topic_number = None
                node = None
                conversation = []
                records = []

                for (
                    query
                ) in (
                    self.dataset.dataset.queries_iter()
                ):  # type: _irds.trec_cast.Cast2020Query
                    decontextualized = DecontextualizedDictItem(
                        "manual",
                        {
                            "manual": query.manual_rewritten_utterance,
                            "auto": query.automatic_rewritten_utterance,
                        },
                    )
                    topic = Cast2020TopicRecord(
                        IDItem(query.query_id),
                        SimpleTextItem(query.raw_utterance),
                        decontextualized,
                        ConversationHistoryItem(
                            node.conversation(False) if node else []
                        ),
                    )

                    if topic_number == query.topic_number:
                        node = node.add(ConversationTreeNode(topic))
                    else:
                        conversation = []
                        node = ConversationTreeNode(topic)
                        topic_number = query.topic_number

                    records.append(topic)

                    conversation.append(node)
                    node = node.add(
                        ConversationTreeNode(
                            Cast2020ResponseRecord(
                                RetrievedEntry(query.manual_canonical_result_id)
                            )
                        )
                    )
                    conversation.append(node)
            except Exception:
                logging.exception("Error while computing topic records")
                raise

            return records

    Topics.HANDLERS.update(
        {
            # _irds.trec_cast.Cast2019Query: Cast2019TopicsHandler,
            _irds.trec_cast.Cast2020Query: Cast2020TopicsHandler,
            # _irds.trec_cast.Cast2021Query: Cast2021TopicsHandler,
            # _irds.trec_cast.Cast2022Query: Cast2022TopicsHandler
        }
    )


class Adhoc(ir.Adhoc, IRDSId):
    pass


class AdhocRun(ir.AdhocRun, IRDSId):
    pass


class TrainingTriplets(ir.TrainingTriplets, IRDSId):
    """Training triplets from IR Dataset"""

    CONVERTERS = {
        GenericDocPair: lambda qid, doc1_id, doc2_id: (
            IDTopicRecord.from_id(qid),
            IDDocumentRecord.from_id(doc1_id),
            IDDocumentRecord.from_id(doc2_id),
        )
    }

    @cached_property
    def topic_recordtype(self) -> Type[Record]:
        """The set of records for topics"""
        return TopicRecord.from_types("TopicIDRecord", IDItem, module=__name__)

    @cached_property
    def document_recordtype(self) -> Type[Record]:
        """The class for documents"""
        return DocumentRecord.from_types("DocumentIDRecord", IDItem, module=__name__)

    @cached_property
    def converter(self):
        return TrainingTriplets.CONVERTERS[self.dataset.docpairs_cls()]

    def iter(
        self,
    ) -> Iterator[Tuple[ir.TopicRecord, ir.DocumentRecord, ir.DocumentRecord]]:
        ds = self.dataset

        logging.info("Starting to generate triplets")
        yield from (self.converter(*entry) for entry in ds.docpairs_iter())
        logging.info("Ending triplet generation")

    def count(self):
        """Returns the length or None"""
        return self.dataset.docpairs_count()



from functools import cached_property
from typing import ClassVar, Tuple
from attrs import define
from datamaestro.record import recordtypes
from ir_datasets.datasets.wapo import WapoDocMedia
from .base import TextItem, SimpleTextItem, IDTopicRecord
from ir_datasets.datasets.cord19 import Cord19FullTextSection


@define
class DocumentWithTitle(TextItem):
    """Web document with title and body"""

    body: str

    title: str

    @cached_property
    def text(self):
        return f"{self.title} {self.body}"


@define
class CordDocument(DocumentWithTitle):
    url: str
    pubmed_id: str


@define
class CordFullTextDocument(TextItem):
    title: str
    doi: str
    date: str
    abstract: str
    body: Tuple[Cord19FullTextSection, ...]

    @cached_property
    def text(self):
        return self.abstract


@define
class MsMarcoDocument(TextItem):
    url: str
    title: str
    body: str

    @cached_property
    def text(self):
        return self.body


@define
class NFCorpusDocument(TextItem):
    url: str
    title: str
    abstract: str

    @cached_property
    def text(self):
        return self.abstract


@define
class TitleDocument(TextItem):
    body: str
    title: str

    @cached_property
    def text(self):
        return f"{self.title} {self.body}"


@define
class TitleUrlDocument(TitleDocument):
    url: str


@define
class TrecParsedDocument(TextItem):
    title: str
    body: str
    marked_up_doc: bytes

    @cached_property
    def text(self):
        return f"{self.title} {self.body}"


@define
class WapoDocument(TextItem):
    url: str
    title: str
    author: str
    published_date: int
    kicker: str
    body: str
    body_paras_html: Tuple[str, ...]
    body_media: Tuple[WapoDocMedia, ...]

    @cached_property
    def text(self):
        return self.body


@define
class TweetDoc(TextItem):
    text: str
    user_id: str
    created_at: str
    lang: str
    reply_doc_id: str
    retweet_doc_id: str
    source: bytes
    source_content_type: str


@define
class OrConvQADocument(TextItem):
    id: str
    title: str
    body: str
    aid: str
    bid: int

    @cached_property
    def text(self):
        return f"{self.title} {self.body}"


@define
class TrecTopic(TextItem):
    text: str
    query: str
    narrative: str


@define
class UrlTopic(TextItem):
    text: str
    url: str


@define
class NFCorpusTopic(TextItem):
    text: str
    all: str


@define
class TrecMb13Query(TextItem):
    query: str
    time: str
    tweet_time: str

    def get_text(self):
        return f"{self.query}"


@define
class TrecMb14Query(TextItem):
    query: str
    time: str
    tweet_time: str
    description: str

    def get_text(self):
        return f"{self.query}"


@define()
class TrecTopic(SimpleTextItem):
    description: str
    narrative: str


@recordtypes(TrecTopic)
class TrecTopicRecord(IDTopicRecord):
    ...



from typing import Iterator, List, Optional
from attr import define
import json
from datamaestro.data import File
from datamaestro.record import recordtypes
from datamaestro_text.data.ir.base import GenericTopicRecord
from .base import (
    AnswerEntry,
    ConversationTree,
    RetrievedEntry,
    SingleConversationTree,
    SimpleDecontextualizedItem,
    AnswerConversationRecord,
)
from . import ConversationDataset


@define(kw_only=True)
class CanardConversation:
    """A query with past history"""

    history: List[str]
    """The list of queries asked by the user"""

    query: str
    """The last issued query"""

    rewrite: str
    """Manually rewritten query"""

    dialogue_id: str
    """Conversation identifier"""

    query_no: int
    """Question number"""


@recordtypes(SimpleDecontextualizedItem)
class CanardTopicRecord(GenericTopicRecord):
    pass


@recordtypes(AnswerEntry, RetrievedEntry)
class CanardAnswerRecord(AnswerConversationRecord):
    pass


class CanardDataset(ConversationDataset, File):
    """A dataset in the CANARD JSON format"""

    def entries(self) -> Iterator[CanardConversation]:
        """Iterates over re-written query with their context"""
        with self.path.open("rt") as fp:
            data = json.load(fp)

        for entry in data:
            yield CanardConversation(
                history=entry["History"],
                query=entry["Question"],
                rewrite=entry["Rewrite"],
                dialogue_id=entry["QuAC_dialog_id"],
                query_no=entry["Question_no"],
            )

    def __iter__(self) -> Iterator[ConversationTree]:
        history = []
        current_id = None

        for entry in self.entries():
            # Check if current conversation
            if current_id != entry.dialogue_id and current_id is not None:
                history.reverse()
                yield SingleConversationTree(current_id, history)

            # Add to current
            history.append(
                # FIXME: not working anymore
                CanardEntry(
                    query=entry.query,
                    decontextualized_query=entry.rewrite,
                )
            )

        yield current



from csv import DictReader
from typing import Iterator

from datamaestro.data import File, documentation
from datamaestro.record import recordtypes
from datamaestro_text.data.ir import Documents, TopicRecord, Topics, IDItem
from datamaestro_text.data.ir.formats import (
    DocumentWithTitle,
    TrecTopicRecord,
    TrecTopic,
)
from datamaestro.data.csv import Generic as GenericCSV
import xml.etree.ElementTree as ET
from datamaestro_text.data.ir.base import GenericDocumentRecord


@recordtypes(DocumentWithTitle)
class CordDocumentRecord(GenericDocumentRecord):
    pass


class Topics(Topics, File):
    """XML format used in Adhoc topics"""

    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        tree = ET.parse(self.path)
        for topic in tree.findall("topic"):
            yield TrecTopicRecord(
                IDItem(topic.get("number")),
                TrecTopic(
                    topic.find("query").text,
                    question=topic.find("question").text,
                    narrative=topic.find("narrative").text,
                ),
            )

    @property
    def topic_recordtype(self):
        return TrecTopicRecord


class Documents(Documents, GenericCSV):
    @documentation
    def iter(self) -> Iterator[CordDocumentRecord]:
        """Returns an iterator over adhoc documents"""
        with self.path.open("r") as fp:
            for row in DictReader(fp):
                yield CordDocumentRecord(
                    IDItem(row["cord_uid"]),
                    DocumentWithTitle(row["abstract"], row["title"]),
                )
